/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

package com.myriad.tmap.compute;

import java.util.Arrays;
import java.io.File;
import java.util.regex.*;
import javax.swing.JOptionPane;

/** This class provides a number of computational methods on a pedigree. Most of
    the work is done by native methods written in C. The Pedigree class provides
    wrappers for these functions and also a cache of results from the computeMap
    method.
 */
public class Pedigree {
    /** Creates a new class from the pedigree data found in the specified file.
     */
    public Pedigree(String aPath) throws Exception
    {
        String[]        emptyKey = new String[0];
        Map             emptyMap = new Map(new MarkerPosition[0]);
        CacheElement    emptyCacheElem = new CacheElement(emptyKey, false,
                                                          emptyMap);
        
        ped = readPedigreeFile(aPath);
        if (ped == 0)
            throw new Exception("Couldn't read pedigree file.");
        cache = new CacheElement[kCacheSize];
        for (int i = 0; i < cache.length; ++i)
            cache[i] = emptyCacheElem;
    }
    
    /** Frees the memory held by the native methods. */
    protected void finalize()
    {
        if (ped != 0)
            freePed(ped);
    }
    
    /** Computes all the two point LOD scores and distances. */
    public TwoPointData compute2Pts()
    {
        int         numMarks = getNumMarkers(ped);
        String[]    names = new String[numMarks];
        double[][]  lods = new double[numMarks][];
        double[][]  dists = new double[numMarks][];
        
        compute2PtsNative(ped, names, lods, dists);
        return new TwoPointData(names, lods, dists);
    }
    
    /** Computes the map (i.e. distances and errors) for the given marker
        order.
     */
    public Map computeMap(String[] markerNames, boolean split)
    {
        for (int i = 0; i < cache.length; ++i) {
            if (cache[i].split == split) {
                Map     map = null;
                
                if (Arrays.equals(cache[i].markerNames, markerNames))
                    map = cache[i].result;
                else if (reverseEquals(cache[i].markerNames, markerNames))
                    map = cache[i].result.reverse();
                
                if (map != null) {
                    /* This means that a reversed map replaces the original. */
                    cacheMap(markerNames, split, map, i);
                    return map;
                }
            }
        }
        
        return makeMap(markerNames, split, MAKE_MAP_COMPUTE_DISTS);
    }
    
    /** Adds a map to the cache.
        @param markerNames the array of markers which make up the map
        @param split whether or not the map was built with separate distances
        for the mother and the father.
        @param map the resulting map.
     */
    private void cacheMap(String[] markerNames, boolean split, Map map)
    {
        cacheMap(markerNames, split, map, cache.length - 1);
    }
    
    /** Adds the specified pairing of markerNames and map to the cache,
        replacing the cache element at the repNdx.
        @param markerNames the array of markers which make up the map
        @param split whether or not the map was built with separate distances
        for the mother and the father.
        @param map the resulting map.
        @param repNdx the index of the cache element to remove to make room for
        the current cache element.
     */
    private void cacheMap(String[] markerNames, boolean split, Map map,
                          int repNdx)
    {
        for (int i = repNdx; i > 0; --i)
            cache[i] = cache[i - 1];
        cache[0] = new CacheElement(markerNames, split, map);
    }
    
    /** Returns a set of markers from the given file. */
    public String[] getOrder(String orderPath)
    {
        return getOrderNative(ped, orderPath);
    }
    
    /** Builds an initial map order based on the markers in the markerNames
        array. On return, markerNames has been altered to reflect the new
        order of the markers.
     */
    public Map build(String[] markerNames, boolean split)
    {
        return makeMap(markerNames, split, MAKE_MAP_BUILD);
    }

    /** Improves the map order indicated by markerNames. On return, the
        markerNames array has been altered to reflect the new order.
     */
    public Map improve(String[] markerNames, boolean split)
    {
        return makeMap(markerNames, split, MAKE_MAP_IMPROVE);
    }
    
    /** Outputs the map for a given marker order to File.
        @param markerNames an array containg the markers which make up the map
        @param split whether the maternal and paternal distances should be
        computed independently
        @param file the file which will contain the output
     */
    public void writeMap(String[] markerNames, boolean split, File file)
    {
        writeMapNative(ped, markerNames, split, file.getAbsolutePath());
    }
    
    /** This is a wrapper function around makeMapNative. It's main purpose is to
        give arrays to the native function and then to package these into the
        Map object.
     */
    private Map makeMap(String[] markerNames, boolean split, int type)
    {
        if (markerNames.length > 0) {
            double[]    distsMoth = new double[markerNames.length - 1];
            double[]    distsFath = new double[markerNames.length - 1];
            double[]    errorsMoth = new double[markerNames.length];
            double[]    errorsFath = new double[markerNames.length];
            Map         map;
            Thread      curThr = Thread.currentThread();
            int         pri = curThr.getPriority();
            
            /* Since this may take a while, we decrease our priority so that
               other work may continue. Compute dists is not affected, since it
               is usually fast and should be interactive.
             */
            if (type != MAKE_MAP_COMPUTE_DISTS)
                curThr.setPriority(Thread.MIN_PRIORITY);
            
            makeMapNative(ped, markerNames, split, distsMoth, distsFath,
                          errorsMoth, errorsFath, type);
            map = buildMap(markerNames, distsMoth, distsFath, errorsMoth,
                           errorsFath);
            
            cacheMap(markerNames, split, map);
            
            if (type != MAKE_MAP_COMPUTE_DISTS)
                curThr.setPriority(pri);
            return map;
        } else
            return new Map(new MarkerPosition[0]);
    }
    
    /** Builds a map class from the given data. The native classes return maps
        in the form of separate double arrays for the distances and the errors.
        This translates this data to a more Java-like structure.
     */
    private Map buildMap(String[] markerNames, double[] distsMoth,
                         double[] distsFath, double[] errorsMoth,
                         double[] errorsFath)
    {
        MarkerPosition[]    markers = new MarkerPosition[markerNames.length];
        double              totalDistMoth = 0.0;
        double              totalDistFath = 0.0;
        
        for (int i = 0; i < markerNames.length; ++i) {
            double  distMoth, distFath;
            
            if (i < markerNames.length - 1) {
                distMoth = distsMoth[i];
                distFath = distsFath[i];
            } else
                distMoth = distFath = 0.0;
            
            markers[i] = new MarkerPosition(markerNames[i],
                                            totalDistMoth, totalDistFath,
                                            errorsMoth[i], errorsFath[i]);
            totalDistMoth += distMoth;
            totalDistFath += distFath;
        }
        
        return new Map(markers);
    }
    
    /** Returns true if and only if one array is equal to the reverse of the
        other. */
    private static boolean reverseEquals(String[] a1, String[] a2)
    {
        int     length = a1.length;
        
        if (length != a2.length)
            return false;
        
        for (int i = 0; i < length; ++i) {
            if (!a1[i].equals(a2[length - i - 1]))
                return false;
        }
        
        return true;
    }
    
    private native long     readPedigreeFile(String path);
    private native void     freePed(long ped);
    private native int      getNumMarkers(long ped);
    private native void     compute2PtsNative(long ped, String[] names,
                                              double[][] lods,
                                              double [][] dists);
    private native String[] getOrderNative(long ped, String orderPath);
    private native void     writeMapNative(long ped, String[] markerNames,
                                           boolean split, String path);
    private native void     makeMapNative(long ped, String[] markerNames,
                                          boolean split,
                                          double[] distsMoth,
                                          double[] distsFath,
                                          double[] errorMoth,
                                          double[] errorFath, int type);
    
    /** Represents a single element in the cache, i.e. a mapping from a list of
        marker names and a split flag to the resulting map.
     */
    private class CacheElement {
        CacheElement(String[] aMarkerNames, boolean aSplit, Map aResult)
        {
            markerNames = aMarkerNames;
            split = aSplit;
            result = aResult;
        }
        
        String[]    markerNames;
        boolean     split;
        Map         result;
    }
    
    private long    ped;
    private CacheElement[]  cache;
    private final static int    kCacheSize = 8;
    
    private final static int    MAKE_MAP_COMPUTE_DISTS = 0;
    private final static int    MAKE_MAP_BUILD = 1;
    private final static int    MAKE_MAP_IMPROVE = 2;
    
    /** This static code section tries to load the native code parts of the
        class. First, it tries using loadLibrary and the default library path.
        Then, it tries to find the path of the jar file which was used to launch
        the, and looks in that directory.
     */
    static {
        boolean succeeded = false;
        
        try {
            System.loadLibrary("FindDists");
            succeeded = true;
        } catch (UnsatisfiedLinkError e) {}
        
        if (!succeeded) {
            String      cp = System.getProperty("java.class.path");
            String[]    libNames = {"FindDists.dll", "libFindDists.so"};
                /* Separates the path component from the jar file name. Assumes:
                     1. The directory separater is not in \w
                     2. The JAR file name is in \w
                     3. There is only one path in the class path. */
            Matcher     matcher = Pattern.compile("(.*?)\\w*.jar").matcher(cp);
            
            if (matcher.matches()) {
                String      path = matcher.group(1);
                
                for (int i = 0; i < libNames.length; ++i) {
                    try {
                        System.load(path + libNames[i]);
                        succeeded = true;
                        break;
                    } catch(UnsatisfiedLinkError e) {}
                }
            }
        }
        
        if (!succeeded) {
            JOptionPane.showMessageDialog(null,
                                          "Couldn't find FindDists library",
                                          "FindDists library",
                                          JOptionPane.ERROR_MESSAGE);
            System.exit(0);
        }
    }
}
