/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

package com.myriad.tmap.grouping;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import javax.swing.*;
import com.myriad.tmap.compute.*;
import com.myriad.tmap.improve.MapComponent;
import com.myriad.tmap.WorkingDialog;

/** Displays the linkage groups formed by the given pedigree for varying LOD
    thresholds. On the right are the groups defined by the threshold of maxLod,
    and towards there is a visual representation of the coalescing of these
    groups as the LOD threshold decreases to minLod.
 */
public class GroupingComponent extends JComponent implements MouseListener {
    public GroupingComponent(Pedigree aPed, double aMinLod, double aMaxLod,
                             double maxDist)
    {
        WorkingDialog   working;
        
        working = new WorkingDialog("Computing 2 point statistics");
        
        minLod = aMinLod;
        maxLod = aMaxLod;
        
        ped = aPed;
        twoPts = ped.compute2Pts();
        createLinkGroupTree(maxDist);
        rootCl = root.cluster(maxLod);
        rootCl.position(0);
        
        saveChooser = null;
        
        addMouseListener(this);
        
        working.dispose();
    }
    
    /** Sets the minimum lod score */
    public void setMinLod(double aMinLod)
    {
        minLod = aMinLod;
        repaint();
    }
    
    /** Sets the maximum lod score */
    public void setMaxLod(double aMaxLod)
    {
        maxLod = aMaxLod;
        rootCl = root.cluster(maxLod);
        rootCl.position(0);
        repaint();
    }
    
    /** Sets the maximum distance between linked markers */
    public void setMaxDist(double maxDist)
    {
        createLinkGroupTree(maxDist);
        rootCl = root.cluster(maxLod);
        rootCl.position(0);
        repaint();
    }
    
    /** Gets an array containing all currently selected markers. */
    public Marker[] getSelectedMarkers()
    {
        return rootCl.getMarkerArray(true);
    }
    
    /** Unselects all markers */
    public void clearSelectedMarkers()
    {
        rootCl.setSelected(false);
        repaint();
    }
    
    /** Creates a new file containing a list of all the selected markers, with
        one marker name on each line.
     */
    public void saveSelectedMarkers()
    {
        Marker[]        marks = rootCl.getMarkerArray(true);
        
        if (marks.length == 0) {
            noMarkersSelError("save");
            return;
        }
        
        if (saveChooser == null) {
            saveChooser = new JFileChooser();
            saveChooser.setDialogTitle("Save linkage group");
        }
        
        if (saveChooser.showSaveDialog(null) == JFileChooser.APPROVE_OPTION) {
            PrintWriter     fp;
            try {
                fp = new PrintWriter(new FileWriter(saveChooser.getSelectedFile()));
            } catch(IOException err) {
                // :TODO: better error handling here
                System.err.println("IOException");
                return;
            }
            
            for (int i = 0; i < marks.length; ++i)
                fp.println(marks[i].getName());
            
            fp.close();
            if (fp.checkError()) {
                // :TODO: better error handling here.
                System.err.println("Error writing markers.");
                return;
            }
            
            for (int i = 0; i < marks.length; ++i)
                marks[i].setUsed(true);
            clearSelectedMarkers();
            
            ChooserFileIncrementer.increment(saveChooser, false);
        }
    }
    
    /** Creates a new frame containing a MapComponent representing an ordered
        map consiting of the currently selected markers.
     */
    public void makeMapFrame(boolean split)
    {
        Marker[]        marks = getSelectedMarkers();
        String[]        markNames = new String[marks.length];
        MapComponent    mapComp;
        
        if (marks.length == 0) {
            noMarkersSelError("build");
            return;
        }
        
        for (int i = 0; i < marks.length; ++i) {
            markNames[i] = marks[i].getName();
            marks[i].setUsed(true);
        }
        
        ped.build(markNames, split);
        /* :TODO: should pass the map structure along here. */
        mapComp = new MapComponent(ped, markNames, split);
        mapComp.putInNewFrame(getLocationOnScreen());
        
        clearSelectedMarkers();
    }
    
    /** Selects the markers whose names are in the array. Doesn't affect the
        selection of the other markers.
     */
    public int selectMarkersByName(String[] markNames)
    {
        int     res = Integer.MAX_VALUE;
        
        for (int i = 0; i < markNames.length; ++i) {
            int pos = selectMarkerByName(markNames[i]);
            if (pos < res)
                res = pos;
        }
        return res;
    }
    
    /** Selects the marker with the given name, without affecting the selection
        status of the other markers.
     */
    public int selectMarkerByName(String markName)
    {
        Marker  mark = rootCl.findMarkerByName(markName);
        
        if (mark != null) {
            mark.setSelected(true);
            return mark.getTop();
        } else
            return 0;
    }
    
    public void paintComponent(Graphics g)
    {
        Graphics2D  g2 = (Graphics2D) g;
        Rectangle   vis = getVisibleRect();
        
        g2.setPaint(Color.white);
        g2.fill(g2.getClipBounds());
        
        g2.setStroke(new BasicStroke(2.0f));
        g2.setFont(new Font("SansSerif", Font.PLAIN, 10));
        
        rootCl.paint(g2, makeDisplayParameters(), vis.y, vis.y + vis.height);
    }
    
    public Dimension getPreferredSize()
    {
        return new Dimension(100, rootCl.getPreferredHeight());
    }
    
    public void mousePressed(MouseEvent event)
    {
        LinkGroupNode   node = rootCl.findNodeForPoint(event.getPoint(),
                                                       makeDisplayParameters());
        
        if (node != null) {
            node.setSelected(!node.isAllSelected());
            repaint();
        }
    }
    
    public void mouseClicked(MouseEvent event) {}
    public void mouseReleased(MouseEvent event) {}
    public void mouseEntered(MouseEvent event) {}
    public void mouseExited(MouseEvent event) {}
    
    /** Creates the tree of linkage group nodes rooted at root from the two
        point statistics and the distThresh.
        
        The leaves of the tree are individual markers, and the internal nodes
        are the lod thresholds at which the branches are linked.
     */
    private void createLinkGroupTree(double distThresh)
    {
        ArrayList       pairsList = new ArrayList();
        Pairing[]       pairsArray;
        ArrayList       nodes = new ArrayList();
        int             numMarks = twoPts.getNumMarkers();
        
        for (int i = 0; i < numMarks; ++i) {
            Marker  curMarker = new Marker(twoPts.getMarkerName(i));
            nodes.add(curMarker);
            
            for (int j = 0; j < i; ++j) {
                if (twoPts.getDist(i, j) < distThresh)
                    pairsList.add(new Pairing(curMarker,
                                              (Marker) nodes.get(j),
                                              twoPts.getLod(i, j)));
            }
        }
        
        pairsArray = (Pairing []) pairsList.toArray(new Pairing[0]);
        Arrays.sort(pairsArray);
        
        for (int i = 0; i < pairsArray.length; ++i) {
            Pairing         pair = pairsArray[i];
            LinkGroupNode   t1 = pair.marker1.getRoot();
            LinkGroupNode   t2 = pair.marker2.getRoot();
            
            if (t1 == t2)
                continue;
            
            for (int j = 0; j < nodes.size();) {
                if (nodes.get(j) == t1 || nodes.get(j) == t2)
                    nodes.remove(j);
                else
                    ++j;
            }
            
            nodes.add(new LinkGroupMerge(t1, t2, pair.lodScore));
        }
        
        root = (LinkGroupNode) nodes.get(0);
        for (int i = 1; i < nodes.size(); ++i)
            root = new LinkGroupMerge(root, (LinkGroupNode) nodes.get(i), 0.0);
    }
    
    /** Creates an appropriate DisplayParameters object for the current
        settings and component size.
     */
    private DisplayParameters makeDisplayParameters()
    {
        DisplayParameters   params = new DisplayParameters();
        
        params.markerX = getSize().width - 100;
        params.minLod = minLod < maxLod ? minLod : maxLod;
        params.maxLod = maxLod;
        params.scale = params.markerX / (params.maxLod - params.minLod);
        
        return params;
    }
    
    private void noMarkersSelError(String action)
    {
        JOptionPane.showMessageDialog(null,
                                      "Can't " + action
                                          + " with no markers selected",
                                      "No markers selected",
                                      JOptionPane.ERROR_MESSAGE);
    }
    
    private Pedigree        ped;
    private TwoPointData    twoPts;
    private LinkGroupNode   root, rootCl;
    private double          minLod;
    private double          maxLod;
    private JFileChooser    saveChooser;
    
    /** Represents linkage between a given pair of markers. */
    private static class Pairing implements Comparable
    {
        Pairing(Marker m1, Marker m2, double lod)
        {
            marker1 = m1;
            marker2 = m2;
            lodScore = lod;
        }
        
        /** Compares two Pairing objects so that they are sorted in order of
            decreasing lod score.
         */
        public int compareTo(Object o)
        {
            Pairing p = (Pairing) o;
            
            if (lodScore < p.lodScore)
                return 1;
            else if (lodScore > p.lodScore)
                return -1;
            else
                return 0;
        }
        
        Marker  marker1;
        Marker  marker2;
        double  lodScore;
    }
}
