/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

package com.myriad.tmap.improve;

import com.myriad.tmap.compute.*;

/** This is used to create a worker thread which computes maps from marker
    orders. The marker orders can be submitted and the results will be returned
    asynchronously to the designated receiver. At any one time there is at most
    one order on the queue. Newer orders always replace the older ones.
 */
class PedigreeThread implements Runnable {
    PedigreeThread(MapReceivable aRecv, Pedigree aPed)
    {
        running = true;
        recv = aRecv;
        ped = aPed;
        order = null;
    }
    
    /** This puts a given map on the queue to be computed. If there is no other
        computation in process, the worker thread will begin computing it
        immediately. Otherwise, it will do so after it has finished with the
        current computation.
     */
    public synchronized void computeMap(String [] aOrder, boolean aSplit)
    {
        order = aOrder;
        split = aSplit;
        notifyAll();
    }
    
    /** This waits for an event from an outside thread, which can be either
        setting running to false in stop(), or setting a new order in
        computeMap(...). In the former case, the method returns null, and in the
        latter it returns the method.
     */
    private synchronized String[] getOrder()
    {
        String[]    result;
        
        while (running) {
            if (order != null) {
                result = order;
                order = null;
                return result;
            }
            
            try {
                wait();
            } catch (InterruptedException e) {}
        }
        
        return null;
    }
    
    /** The working thread loop. It calls getOrder to get an event, either a
        stop or a map order. In the latter case it then computes the map.
     */
    public void run()
    {
        for(;;) {
            String[]    order = getOrder();
            Map         map;
            
            if (order == null)
                break;
            recv.mapResult(ped.computeMap(order, split));
        }
    }
    
    /** Stops the thread by setting running equal to false.
     */
    public synchronized void stop()
    {
        running = false;
        notifyAll();
    }
    
    private boolean         running;
    private MapReceivable   recv;
    private Pedigree        ped;
    private String[]        order;
    private boolean         split;
};
